# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vizcode']

package_data = \
{'': ['*'],
 'vizcode': ['frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-192x192.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/android-chrome-512x512.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/apple-touch-icon.png',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/asset-manifest.json',
             'frontend/build/code/*',
             'frontend/build/code/python/*',
             'frontend/build/code/python/main/.gitkeep',
             'frontend/build/code/python/main/edge-code/.gitkeep',
             'frontend/build/code/python/main/node-code/.gitkeep',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/company-icon.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-16x16.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon-32x32.png',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/favicon.ico',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/index.html',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/manifest.json',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/robots.txt',
             'frontend/build/static/css/*',
             'frontend/build/static/js/*']}

install_requires = \
['art>=5.1,<6.0',
 'jedi>=0.17.2,<0.18.0',
 'parso>=0.7.0,<0.8.0',
 'pickleDB>=0.9.2,<0.10.0']

entry_points = \
{'console_scripts': ['vizcode = vizcode.main:start']}

setup_kwargs = {
    'name': 'vizcode',
    'version': '0.3.10',
    'description': 'Visualize your codebase as a graph',
    'long_description': '# VizCode\n\nVizCode is a command line tool for visualizing codebases as directed graphs. It utilizes static analysis to parse through your code to detect any definitions and references, which are then represented as nodes and edges respectively in a graph. Furthermore, the graph will be visualized on a web browser once the code files have been parsed. VizCode only supports the Python language at the moment.\n\n![](https://i.ibb.co/28S6ZFx/vizcode-screenshot.png)\n\nPlease note that VizCode is run locally and will never send your code, repository names, file names, or any other specific code data to any location outside of your local machine. Additionally, VizCode is not able to run on Windows unless you download the Ubuntu shell. For more information about VizCode visit our [website](https://vizcode.dev).\n\n## Installation\n\nYou can install VizCode from [PyPI](https://pypi.org/project/vizcode/):\n\n    pip install vizcode\n\nVizCode is supported on Python 3.7 and above.\n\n## How to use\n\nTo run VizCode on a codebase, simply call the program:\n\n    $ vizcode -p [ENTER A DIRECTORY PATH]\n\nSince VizCode only works on Python code, it will ignore any files without the .py file extension. If your code uses any dependencies that are not installed in your default environment, then you can specify an environment path:\n\n    $ vizcode -p [ENTER A DIRECTORY PATH] -e [ENTER AN ENVIRONMENT PATH]\n\nIf you would like VizCode to ignore test files, then you can add an additional argument:\n\n    $ vizcode -p [ENTER A DIRECTORY PATH] -t [ENTER PATH OF TEST FILES]\n',
    'author': 'Alex Rubin, Maciej Holubiec, Karthik Rao',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://vizcode.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
